/*
 * U-boot - flash.c Flash driver for SST39VF3201
 *
 * Copyright (c) 2017 Kaneko System Co., Ltd.
 * Copyright (c) 2005-2007 Analog Devices Inc.
 * This file is based on BF533EzFlash.c originally written by Analog Devices, Inc.
 *
 * (C) Copyright 2000-2004
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 * MA 02110-1301 USA
 */

#include <asm/io.h>
#include "flash-defines.h"

flash_info_t flash_info[CONFIG_SYS_MAX_FLASH_BANKS];

static int get_codes(void);
static int poll_toggle_bit(long lOffset);
static void exit_flash(void);
static int erase_flash(void);
static int erase_block_flash(int, unsigned long);
static void unlock_flash(long lOffset);
static int write_data(long lStart, long lCount, uchar *pnData);
static int read_data(long lStart, long lCount, long lStride, int *pnData);
static int read_flash(long nOffset, int *pnValue);
static int write_flash(long nOffset, int nValue);
static void get_sector_number(long lOffset, int *pnSector);

void flash_reset(void)
{
	exit_flash();
}

unsigned long flash_get_size(ulong baseaddr, flash_info_t * info)
{
	int id = 0, i = 0;
	static int FlagDev = 1;

	id = get_codes();
	if (FlagDev) {
#ifdef DEBUG
		printf("Device ID of the Flash is %x\n", id);
#endif
		FlagDev = 0;
	}
	info->flash_id = id;

	for (i = 0; i < NUM_SECTORS; i++)
		info->start[i] = (baseaddr + (i * SECTOR_SIZE));
	info->size = FLASH_SIZE;
	info->sector_count = NUM_SECTORS;
	return (info->size);
}

unsigned long flash_init(void)
{
	unsigned long size_b0;

	size_b0 = 0;
#ifdef DEBUG
	printf("Flash Memory Start 0x%x\n", CONFIG_SYS_FLASH_BASE);
	printf("Memory Map for the Flash\n");
	printf("0x20000000 - 0x200FFFFF Flash (1MB)\n");
	printf("Please type command flinfo for information on Sectors \n");
#endif
	flash_info[0].flash_id = FLASH_UNKNOWN;
	size_b0 = flash_get_size(CONFIG_SYS_FLASH_BASE, &flash_info[0]);

	if (flash_info[0].flash_id == FLASH_UNKNOWN || size_b0 == 0) {
		printf("## Unknown FLASH on Bank 0 - Size = 0x%08lx = %ld MB\n",
		       size_b0, size_b0 >> 20);
	}

	return size_b0;
}

void flash_print_info(flash_info_t * info)
{
	int i;

	if (info->flash_id == FLASH_UNKNOWN) {
		printf("missing or unknown FLASH type\n");
		return;
	}

	switch (info->flash_id) {
	case 0x235B:
		printf("ST Microelectronics SST39VF3202");
		break;
	default:
		printf("Unknown Vendor: (0x%08lX) ", info->flash_id);
		break;
	}
	for (i = 0; i < info->sector_count; ++i) {
		if ((i % 5) == 0)
			printf("\n   ");
		printf(" %08lX%s",
		       info->start[i], info->protect[i] ? " (RO)" : "     ");
	}
	printf("\n");
	return;
}

int flash_erase(flash_info_t * info, int s_first, int s_last)
{
	int i;
	int prot, sect;

	prot = 0;
	for (sect = s_first; sect <= s_last; ++sect) {
		if (info->protect[sect])
			prot++;
	}

	if (prot)
		printf("- Warning: %d protected sectors will not be erased!\n",
		       prot);
	else
		printf("\n");

	printf("Erasing Flash locations, Please Wait\n");
	for (i = s_first; i <= s_last; i++) {
		if (info->protect[i] == 0) {	/* not protected */
			if (erase_block_flash(i, info->start[i]) < 0) {
				printf("Error Sector erasing \n");
				return FLASH_FAIL;
			}
		}
	}
	return FLASH_SUCCESS;
}

int write_buff(flash_info_t * info, uchar * src, ulong addr, ulong cnt)
{
	int ret;
	int d;
	if (addr % 2) {
		read_flash(addr - 1 - CONFIG_SYS_FLASH_BASE, &d);
		d = (int)((d & 0x00FF) | (*src++ << 8));
		ret = write_data(addr - 1, 2, (uchar *) & d);
		if (ret == FLASH_FAIL)
			return ERR_NOT_ERASED;
		ret = write_data(addr + 1, cnt - 1, src);
	} else
		ret = write_data(addr, cnt, src);
	if (ret == FLASH_FAIL)
		return ERR_NOT_ERASED;
	return FLASH_SUCCESS;
}

int write_data(long lStart, long lCount, uchar * pnData)
{
	long i = 0;
	unsigned long ulOffset = lStart - CONFIG_SYS_FLASH_BASE;
	int d;
	int nSector = 0;
	int flag = 0;

	if (lCount % 2) {
		flag = 1;
		lCount = lCount - 1;
	}

	for (i = 0; i < lCount - 1; i += 2, ulOffset += 2) {
		get_sector_number(ulOffset, &nSector);
		read_flash(ulOffset, &d);
		if (d != 0xffff) {
			printf
			    ("Flash not erased at offset 0x%lx Please erase to reprogram\n",
			     ulOffset);
			return FLASH_FAIL;
		}
		unlock_flash(ulOffset);
		d = (int)(pnData[i] | pnData[i + 1] << 8);
		write_flash(WRITESEQ1, WRITEDATA1);
		write_flash(WRITESEQ2, WRITEDATA2);
		write_flash(WRITESEQ3, WRITEDATA3);
		write_flash(ulOffset, d);
		if (poll_toggle_bit(ulOffset) < 0) {
			printf("Error programming the flash \n");
			return FLASH_FAIL;
		}
		if ((i > 0) && (!(i % SECTOR_SIZE)))
			printf(".");
	}
	if (flag) {
		get_sector_number(ulOffset, &nSector);
		read_flash(ulOffset, &d);
		if (d != 0xffff) {
			printf
			    ("Flash not erased at offset 0x%lx Please erase to reprogram\n",
			     ulOffset);
			return FLASH_FAIL;
		}
		unlock_flash(ulOffset);
		d = (int)(pnData[i] | (d & 0xFF00));
		write_flash(WRITESEQ1, WRITEDATA1);
		write_flash(WRITESEQ2, WRITEDATA2);
		write_flash(WRITESEQ3, WRITEDATA3);
		write_flash(ulOffset, d);
		if (poll_toggle_bit(ulOffset) < 0) {
			printf("Error programming the flash \n");
			return FLASH_FAIL;
		}
	}
	return FLASH_SUCCESS;
}

int read_data(long ulStart, long lCount, long lStride, int *pnData)
{
	long i = 0;
	int j = 0;
	long ulOffset = ulStart;
	int iShift = 0;
	int iNumWords = 2;
	int nLeftover = lCount % 4;
	int nHi, nLow;
	int nSector = 0;

	for (i = 0; (i < lCount / 4) && (i < BUFFER_SIZE); i++) {
		for (iShift = 0, j = 0; j < iNumWords; j += 2) {
			get_sector_number(ulOffset, &nSector);
			read_flash(ulOffset, &nLow);
			ulOffset += (lStride * 2);
			read_flash(ulOffset, &nHi);
			ulOffset += (lStride * 2);
			pnData[i] = (nHi << 16) | nLow;
		}
	}
	if (nLeftover > 0) {
		get_sector_number(ulOffset, &nSector);
		read_flash(ulOffset, &pnData[i]);
	}
	return FLASH_SUCCESS;
}

int write_flash(long nOffset, int nValue)
{
	long addr;

	addr = (CONFIG_SYS_FLASH_BASE + nOffset);
	SSYNC();
	*(unsigned volatile short *)addr = nValue;
	SSYNC();
/*
	if (poll_toggle_bit(nOffset) < 0)
		return FLASH_FAIL;
*/
	return FLASH_SUCCESS;
}

int read_flash(long nOffset, int *pnValue)
{
	int nValue = 0x0;
	long addr = (CONFIG_SYS_FLASH_BASE + nOffset);

	nValue = *(volatile unsigned short *)addr;
	SSYNC();
	*pnValue = nValue;
	return true;
}

int poll_toggle_bit(long lOffset)
{
	unsigned int u1, u2;
	unsigned long timeout = 0xFFFFFFFF;
	volatile unsigned long *FB =
	    (volatile unsigned long *)(0x20000000 + lOffset);
	while (1) {
		if (timeout < 0)
			break;
		u1 = *(volatile unsigned short *)FB;
		u2 = *(volatile unsigned short *)FB;
		if ((u1 & 0x0040) == (u2 & 0x0040))
			return FLASH_SUCCESS;
		timeout--;
	}
	printf("Time out occured \n");
	if (timeout < 0)
		return FLASH_FAIL;
}

void exit_flash(void)
{
	write_flash(WRITESEQ1, EXITENTRY1);
	write_flash(WRITESEQ2, EXITENTRY2);
	write_flash(WRITESEQ2, EXITENTRY3);
}

int erase_flash(void)
{
	int i;

	for(i = 0; i < NUM_SECTORS; i++)
	{
		flash_erase(&flash_info, 0, NUM_SECTORS);
	}
	return FLASH_SUCCESS;
}

int erase_block_flash(int nBlock, unsigned long address)
{
	long ulSectorOff = 0x0;

	if ((nBlock < 0) || (nBlock > NUM_SECTORS))
		return false;

	ulSectorOff = (address - CONFIG_SYS_FLASH_BASE);

	write_flash(WRITESEQ1, ERASEDATA1);
	write_flash(WRITESEQ2, ERASEDATA2);
	write_flash(WRITESEQ3, ERASEDATA3);
	write_flash(WRITESEQ4, ERASEDATA4);
	write_flash(WRITESEQ5, ERASEDATA5);
	write_flash(ulSectorOff, ERASEDATA6);

	if (poll_toggle_bit(ulSectorOff) < 0)
		return FLASH_FAIL;

	return FLASH_SUCCESS;
}

void unlock_flash(long ulOffset)
{
}

int get_codes()
{
	int dev_id = 0;

	write_flash(WRITESEQ1, GETCODEDATA1);
	write_flash(WRITESEQ2, GETCODEDATA2);
	write_flash(WRITESEQ3, GETCODEDATA3);

	read_flash(0x0002, &dev_id);
	dev_id &= 0xFFFF;

	exit_flash();

	return dev_id;
}

void get_sector_number(long ulOffset, int *pnSector)
{
	*pnSector = ulOffset / SECTOR_SIZE;
}

